import numpy as np
import pandas as pd
from mne import Epochs, events_from_annotations
from mne.time_frequency import tfr_morlet

from .averaging import compute_evokeds
from .epoching import (compute_single_trials, get_bad_channels, get_bad_epochs,
                       match_log_to_epochs, read_log, triggers_to_event_id)
from .io import (read_raw, save_clean, save_df, save_epochs, save_evokeds,
                 save_montage, save_report)
from .preprocessing import (add_heog_veog, apply_montage, correct_besa,
                            correct_ica, interpolate_bad_channels)
from .report import create_report
from .tfr import compute_single_trials_tfr, subtract_evoked


def participant_pipeline(
    vhdr_file,
    log_file,
    besa_file=None,
    bad_channels=None,
    auto_bad_channels=None,
    skip_log_rows=None,
    skip_log_conditions=None,
    downsample_sfreq=None,
    veog_channels='auto',
    heog_channels='auto',
    montage='easycap-M1',
    ica_method=None,
    ica_n_components=0.99,
    highpass_freq=0.1,
    lowpass_freq=40.0,
    triggers=None,
    triggers_column=None,
    epochs_tmin=-0.5,
    epochs_tmax=1.5,
    baseline=(-0.2, 0.0),
    reject_peak_to_peak=200.0,
    components={'name': [], 'tmin': [], 'tmax': [], 'roi': []},
    average_by=None,
    perform_tfr=False,
    tfr_subtract_evoked=False,
    tfr_freqs=np.linspace(4.0, 40.0, num=37),
    tfr_cycles=np.linspace(2.0, 20.0, num=37),
    tfr_mode='percent',
    tfr_baseline=(-0.45, -0.05),
    tfr_components={
        'name': [], 'tmin': [], 'tmax': [], 'fmin': [], 'fmax': [], 'roi': []},
    clean_dir=None,
    epochs_dir=None,
    trials_dir=None,
    evokeds_dir=None,
    chanlocs_dir=None,
    tfr_dir=None,
    report_dir=None,
    to_df=True,
):
    """Processes EEG data for a single participant.

    The raw data is read and cleaned using standard steps (downsampling, bad
    channel interpolation, ocular correction, frequency domain filtering).
    Epochs are created around the `triggers`. Bad epochs are removed based on
    peak-to-peak amplitude. Single trial mean ERP amplitudes for ERP
    `components` of interest are computed and matched to the single trial
    behavioral data from the `log_file`.

    Optionally, this last step is repeated on a time-frequency representation
    (TFR) of the data obtained via Morlet wavelet convolution.

    The result is a single trial data frame which can be used for fitting
    linear mixed-effects models on the mean ERP amplitudes (and power).

    Additionally, condition averages (`evokeds`) for the ERPs (and power) are
    computed to facilitate plotting.

    For details about the pipeline, see Frömer et al. (2018)[1].

    Parameters & returns
    --------------------
    See the README[2] in the GitHub repository for the pipeline.

    Notes
    -----
    [1] https://doi.org/10.3389/fnins.2018.00048
    [2] https://github.com/alexenge/hu-neuro-pipeline/blob/dev/README.md
    """

    # Backup input arguments for re-use
    config = locals()

    # Read raw data
    raw, participant_id = read_raw(vhdr_file)

    # Create backup of the raw data for the HTML report
    if report_dir is not None:
        dirty = raw.copy()

    # Downsample
    if downsample_sfreq is not None:
        sfreq = raw.info['sfreq']
        downsample_sfreq = float(downsample_sfreq)
        print(f'Downsampling from {sfreq} Hz to {downsample_sfreq} Hz')
        raw.resample(downsample_sfreq)

    # Add EOG channels
    raw = add_heog_veog(raw, veog_channels, heog_channels)

    # Apply custom or standard montage
    apply_montage(raw, montage)

    # Handle any bad channels
    raw, interpolated_channels = interpolate_bad_channels(
        raw, bad_channels, auto_bad_channels)

    # Re-reference to common average
    _ = raw.set_eeg_reference('average')

    # Do ocular correction with BESA and/or ICA
    if besa_file is not None:
        raw = correct_besa(raw, besa_file)
    if ica_method is not None:
        raw, ica = correct_ica(raw, ica_method, ica_n_components)
    else:
        ica = None

    # Filtering
    filt = raw.copy().filter(highpass_freq, lowpass_freq, n_jobs=1)

    # Determine events and the corresponding (selection of) triggers
    events, event_id = events_from_annotations(
        filt, regexp='Stimulus', verbose=False)
    if triggers is not None:
        event_id = triggers_to_event_id(triggers)

    # Epoching including baseline correction
    if baseline is not None:
        baseline = tuple(baseline)
    epochs = Epochs(filt, events, event_id, epochs_tmin, epochs_tmax, baseline,
                    preload=True, on_missing='warn')

    # Automatically detect bad channels and interpolate if necessary
    if bad_channels == 'auto' and auto_bad_channels is None:
        auto_bad_channels = get_bad_channels(epochs)
        config['auto_bad_channels'] = auto_bad_channels
        if auto_bad_channels != []:
            print('Restarting with interpolation of bad channels')
            return participant_pipeline(**config)

    # Add bad ICA components to config
    if ica is not None:
        if ica_n_components < 1.0:
            config['auto_ica_n_components'] = int(ica.n_components_)
        config['auto_ica_bad_components'] = [int(x) for x in ica.exclude]

    # Drop the last sample to produce a nice even number
    _ = epochs.crop(tmin=None, tmax=epochs_tmax, include_tmax=False)
    print(epochs.__str__().replace(u"\u2013", "-"))

    # Read behavioral log file and match to the epochs
    log = read_log(log_file, skip_log_rows, skip_log_conditions)
    if triggers_column is not None:
        log, missing_ixs = match_log_to_epochs(epochs, log, triggers_column)
        config['auto_missing_epochs'] = missing_ixs
    epochs.metadata = log
    epochs.metadata.insert(0, column='participant_id', value=participant_id)

    # If log file was provided as a DataFrame, convert to dict for config
    if isinstance(log_file, pd.DataFrame):
        log_file = log_file.astype(object)  # Convert NaNs to null for JSON
        log_file = log_file.where(pd.notnull(log_file), None)
        config['log_file'] = log_file.to_dict(orient='list')

    # Get indices of bad epochs
    bad_ixs = get_bad_epochs(epochs, reject_peak_to_peak)
    config['auto_rejected_epochs'] = bad_ixs

    # Compute single trial mean ERP amplitudes and add to metadata
    trials = compute_single_trials(epochs, components, bad_ixs)

    # Compute evokeds
    evokeds, evokeds_df = compute_evokeds(
        epochs, average_by, bad_ixs, participant_id)

    # Save cleaned continuous data
    if clean_dir is not None:
        save_clean(filt, clean_dir, participant_id)

    # Save channel locations
    if chanlocs_dir is not None:
        save_montage(epochs, chanlocs_dir)

    # Save epochs as data frame and/or MNE object
    if epochs_dir is not None:
        save_epochs(epochs, epochs_dir, participant_id, to_df)

    # Save evokeds as data frame and/or MNE object
    if evokeds_dir is not None:
        save_evokeds(evokeds, evokeds_df, evokeds_dir, participant_id, to_df)

    # Create and save HTML report
    if report_dir is not None:
        dirty.info['bads'] = interpolated_channels
        report = create_report(participant_id, dirty, ica, filt, events,
                               event_id, epochs, evokeds)
        save_report(report, report_dir, participant_id)

    # Time-frequency analysis
    if perform_tfr:

        # Epoching again without filtering
        epochs_unfilt = Epochs(raw, events, event_id, epochs_tmin, epochs_tmax,
                               baseline, preload=True, on_missing='warn',
                               verbose=False)

        # Drop the last sample to produce a nice even number
        _ = epochs_unfilt.crop(tmin=None, tmax=epochs_tmax, include_tmax=False)

        # Copy original metadata
        epochs_unfilt.metadata = epochs.metadata.copy()

        # Optionally subtract evoked activity
        # See, e.g., https://doi.org/10.1016/j.neuroimage.2006.02.034
        if tfr_subtract_evoked:
            epochs_unfilt = subtract_evoked(epochs_unfilt, average_by, evokeds)

        # Morlet wavelet convolution
        print('Doing time-frequency transform with Morlet wavelets')
        tfr = tfr_morlet(epochs_unfilt, tfr_freqs, tfr_cycles, use_fft=True,
                         return_itc=False, n_jobs=1, average=False)

        # First, divisive baseline correction using the full epoch
        # See https://doi.org/10.3389/fpsyg.2011.00236
        if tfr_mode is not None:
            tfr_modes = \
                ['ratio', 'logratio', 'percent', 'zscore', 'zlogratio']
            assert tfr_mode in tfr_modes, \
                f'`tfr_baseline_mode` must be one of {tfr_modes}'
            tfr.apply_baseline(baseline=(None, None), mode=tfr_mode)

        # Second, additive baseline correction using the prestimulus interval
        if tfr_baseline is not None:
            tfr_baseline = tuple(tfr_baseline)
        tfr.apply_baseline(baseline=tfr_baseline, mode='mean')

        # Reduce numerical precision to reduce object size
        tfr.data = np.float32(tfr.data)

        # Add single trial mean power to metadata
        trials = compute_single_trials_tfr(tfr, tfr_components, bad_ixs)

        # Save single trial data (again)
        if trials_dir is not None:
            save_df(trials, trials_dir, participant_id, suffix='trials')

        # Compute evoked power
        tfr_evokeds, tfr_evokeds_df = compute_evokeds(
            tfr, average_by, bad_ixs, participant_id)

        # Save evoked power
        if tfr_dir is not None:
            save_evokeds(
                tfr_evokeds, tfr_evokeds_df, tfr_dir, participant_id, to_df)

        return trials, evokeds, evokeds_df, config, tfr_evokeds, tfr_evokeds_df

    return trials, evokeds, evokeds_df, config



import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from mne import Report, set_log_level


def create_report(
    participant_id, raw, ica, clean, events, event_id, epochs, evokeds):
    """Creates a HTML report for the processing steps of one participant."""

    # Disable warnings about number of open figures
    mpl.rcParams.update({'figure.max_open_warning': 0})
    plt.ioff()

    # Initialize HTML report
    print('Creating HTML report')
    set_log_level('ERROR')
    report = Report(title=f'Report for {participant_id}', verbose=False)

    # Add raw data info
    report.add_raw(raw, title='Raw data', butterfly=False)

    # Add raw time series plots
    n_figs = 10
    raw_figs = plot_time_series(raw, n_figs)
    captions = [f'Segment {i + 1} of {n_figs}' for i in range(n_figs)]
    report.add_figure(
        raw_figs, title='Raw time series', caption=captions, tags=('raw',))

    # Add ICA
    if ica is not None:
        raw.info['bads'] = []  # Else plotting ICA fails
        report.add_ica(ica, title='ICA', inst=raw)

    # Add cleaned data info
    report.add_raw(
        clean, title='Cleaned data', butterfly=False, tags=('clean',))

    # Add cleaned time series plots
    clean_figs = plot_time_series(clean, n_figs)
    report.add_figure(clean_figs, title='Cleaned time series',
                      caption=captions, tags=('clean',))

    # Add events
    sfreq = clean.info['sfreq']
    report.add_events(
        events, title='Event triggers', event_id=event_id, sfreq=sfreq)

    # Add epochs
    report.add_epochs(epochs, title='Epochs')

    # Add evokeds
    report.add_evokeds(evokeds)  # Automatically uses comments as titles
    set_log_level('INFO')

    return report


def plot_time_series(raw, n=10, duration=10.):
    """Plots some seconds of raw data for all channels at `n` time points."""

    # Get evenly spaced starting time points
    starts = np.linspace(0, raw.times[-1], num=n, endpoint=False).astype('int')

    # Create an all-channel plot for each time window
    n_channels = len(raw.ch_names)
    figs = [raw.plot(
        start=start,
        duration=duration,
        n_channels=n_channels,
        bad_color='red',
        show=False,
        show_scrollbars=False,
        show_scalebars=False)
        for start in starts]
    
    return figs



import numpy as np
import pandas as pd
from mne.channels import find_ch_adjacency
from mne.stats import combine_adjacency, permutation_cluster_1samp_test


def compute_perm(evokeds_per_participant, contrasts, tmin=0.0, tmax=1.0,
                 channels=None, n_jobs=1, n_permutations=5001, seed=1234):
    """Performs a cluster based permutation test for a given contrast"""

    # Extract one example evoked for reading data dimensions
    example_evoked = evokeds_per_participant[0][0].copy()

    # Get relevant time samples
    times = example_evoked.times
    if tmin is not None:
        times = [t for t in times if t >= tmin]
    if tmax is not None:
        times = [t for t in times if t < tmax]

    # Get relevant channels
    if channels is None:
        channels = example_evoked.pick_types(eeg=True).ch_names
    else:
        assert all([ch in example_evoked.ch_names for ch in channels]), \
            'All channels in `perm_channels` must be present in the data!'

    # Get dimensions of data for the permutation test
    n_participants = len(evokeds_per_participant)
    n_times = len(times)
    n_channels = len(channels)

    # Prepare emtpy list for results
    cluster_dfs = []

    # Sequentially handle each contrast
    for contrast in contrasts:

        # Prepare empty array
        X = np.zeros((n_participants, n_times, n_channels))

        # Compute a difference wave for each participant
        for ix, evokeds in enumerate(evokeds_per_participant):

            # Extract evoked data for the two conditions of interest
            data_conditions = []
            for condition in contrast:
                evoked = [ev for ev in evokeds if ev.comment == condition][0]
                evoked = evoked.copy().crop(
                    tmin, tmax, include_tmax=False).pick_channels(channels)
                data_conditions.append(evoked.data)

            # Compute difference between conditions
            data_diff = data_conditions[0] - data_conditions[1]
            data_diff = data_diff.swapaxes(1, 0)  # Time points, channels
            X[ix] = data_diff

        # Compute channel adjacency matrix
        ch_adjacency, _ = find_ch_adjacency(evoked.info, 'eeg')

        # Run permutation test
        t_obs, clusters, cluster_p_vals, H0 = permutation_cluster_1samp_test(
            X, n_permutations=n_permutations, adjacency=ch_adjacency,
            n_jobs=n_jobs, seed=seed)

        # Sort clusters by p values
        cluster_ranks = cluster_p_vals.argsort()
        cluster_p_vals = cluster_p_vals[cluster_ranks]
        clusters = [clusters[rank] for rank in cluster_ranks]

        # Create cluster images with cluster labels and p values
        labels = np.full_like(t_obs, 'NA', dtype=object)
        p_vals = np.ones_like(t_obs)
        pos_ix = 0
        neg_ix = 0
        for ix, cluster in enumerate(clusters):

            # Check if the cluster is positive or negative
            if t_obs[cluster][0] > 0:
                pos_ix += 1
                labels[cluster] = f'pos_{pos_ix}'
            else:
                neg_ix += 1
                labels[cluster] = f'neg_{neg_ix}'

            # Extract cluster level p value
            p_val = cluster_p_vals[ix]
            p_vals[cluster] = p_val

        # Prepare DataFrame for storing t values, cluster labels, and p values
        cluster_df = pd.DataFrame({
            'contrast': ' - '.join(contrast),
            'time': np.repeat(times, repeats=n_channels),
            'channel': np.tile(channels, reps=n_times)})

        # Add t values, cluster labels, and p values
        arrs = [t_obs, labels, p_vals]
        stats = ['t_obs', 'cluster', 'p_val']
        for arr, stat in zip(arrs, stats):

            # Convert array to long format
            # Initial array is has shape (times, channels)
            # New array is has shape (times * channels,)
            arr_long = arr.flatten()

            # Add to DataFrame
            cluster_df[stat] = arr_long

        # Append to the list of all contrasts
        cluster_dfs.append(cluster_df)

    # Combine DataFrames of all contrasts
    cluster_df = pd.concat(cluster_dfs, ignore_index=True)

    return cluster_df


def compute_perm_tfr(
        evokeds_per_participant, contrasts, tmin=0.0, tmax=1.0, channels=None,
        fmin=None, fmax=None, n_jobs=1, n_permutations=5001, seed=1234):
    """Performs a cluster based permutation test on time-frequency data"""

    # Extract one example evoked for reading data dimensions
    example_evoked = evokeds_per_participant[0][0].copy()

    # Get relevant time samples
    times = example_evoked.times
    if tmin is not None:
        times = [t for t in times if t >= tmin]
    if tmax is not None:
        times = [t for t in times if t < tmax]

    # Get relevant frequencies
    freqs = example_evoked.freqs
    if fmin is not None:
        freqs = [f for f in freqs if f >= fmin]
    if fmax is not None:
        freqs = [f for f in freqs if f < fmax]

    # Get relevant channels
    if channels is None:
        channels = example_evoked.pick_types(eeg=True).ch_names
    else:
        assert all([ch in example_evoked.ch_names for ch in channels]), \
            'All channels in `perm_channels` must be present in the data!'

    # Get dimensions of data for the permutation test
    n_participants = len(evokeds_per_participant)
    n_times = len(times)
    n_freqs = len(freqs)
    n_channels = len(channels)

    # Prepare emtpy list for results
    cluster_dfs = []

    # Sequentially handle each contrast
    for contrast in contrasts:

        # Prepare empty array
        X = np.zeros((n_participants, n_times, n_freqs, n_channels))

        # Compute a difference wave for each participant
        for ix, evokeds in enumerate(evokeds_per_participant):

            # Extract evoked data for the two conditions of interest
            data_conditions = []
            for condition in contrast:
                evoked = [ev for ev in evokeds if ev.comment == condition][0]
                evoked = evoked.copy().crop(
                    tmin, tmax, fmin, fmax, include_tmax=False).pick_channels(
                        channels)
                data_conditions.append(evoked.data)

            # Compute difference between conditions
            data_diff = data_conditions[0] - data_conditions[1]
            data_diff = data_diff.swapaxes(0, 2)  # Times, freqs, channels
            X[ix] = data_diff

        # Compute frequency and channel adjacency matrix
        # Based on channel locations and a lattice matrix for frequencies
        ch_adjacency, _ = find_ch_adjacency(evoked.info, 'eeg')
        adjacency = combine_adjacency(n_freqs, ch_adjacency)

        # Run permutation test
        t_obs, clusters, cluster_p_vals, H0 = permutation_cluster_1samp_test(
            X, n_permutations=n_permutations, adjacency=adjacency,
            n_jobs=n_jobs, seed=seed)

        # Sort clusters by p values
        cluster_ranks = cluster_p_vals.argsort()
        cluster_p_vals = cluster_p_vals[cluster_ranks]
        clusters = [clusters[rank] for rank in cluster_ranks]

        # Create cluster images with cluster labels and p values
        labels = np.full_like(t_obs, 'NA', dtype=object)
        p_vals = np.ones_like(t_obs)
        pos_ix = 0
        neg_ix = 0
        for ix, cluster in enumerate(clusters):

            # Check if the cluster is positive or negative
            if t_obs[cluster][0] > 0:
                pos_ix += 1
                labels[cluster] = f'pos_{pos_ix}'
            else:
                neg_ix += 1
                labels[cluster] = f'neg_{neg_ix}'

            # Extract cluster level p value
            p_val = cluster_p_vals[ix]
            p_vals[cluster] = p_val

        # Prepare DataFrame for storing t values, cluster labels, and p values
        cluster_df = pd.DataFrame({
            'contrast': ' - '.join(contrast),
            'time': np.repeat(times, n_channels * n_freqs),
            'freq': np.repeat(np.tile(freqs, n_times), n_channels),
            'channel': np.tile(channels, n_times * n_freqs)})

        # Add t values, cluster labels, and p values
        arrs = [t_obs, labels, p_vals]
        stats = ['t_obs', 'cluster', 'p_val']
        for arr, stat in zip(arrs, stats):

            # Convert array to long format
            # Initial array is has shape (times, channels)
            # New array is has shape (times * channels,)
            arr_long = arr.flatten()

            # Add to DataFrame
            cluster_df[stat] = arr_long

        # Append to the list of all contrasts
        cluster_dfs.append(cluster_df)

    # Combine DataFrames of all contrasts
    cluster_df = pd.concat(cluster_dfs, ignore_index=True)

    return cluster_df



from os import path

import pandas as pd
from mne import set_bipolar_reference
from mne.channels import make_standard_montage, read_custom_montage
from mne.preprocessing import ICA


def add_heog_veog(raw, veog_channels='auto', heog_channels='auto'):
    """Adds virtual VEOG and HEOG using default or non-default EOG names."""

    # Add bipolar VEOG channel
    if veog_channels is not None:
        if veog_channels == 'auto':
            veog_channels = ['Fp1', 'FP1', 'Auge_u', 'IO1']
        raw = add_eog(raw, veog_channels, new_name='VEOG')

    # Add bipolar HEOG channel
    if heog_channels is not None:
        if heog_channels == 'auto':
            heog_channels = ['F9', 'F10', 'Afp9', 'Afp10']
        raw = add_eog(raw, heog_channels, new_name='HEOG')

    return raw


def add_eog(raw, channels, new_name):
    """Computes a single bipolar EOG channel from a list of possible names."""

    # Check that exactly two of the provided channels are in the data
    channels = [ch for ch in channels if ch in raw.ch_names]
    assert len(channels) == 2, (
        'Could not find exactly two channels for computing bipolar '
        f'{new_name}. Please provide different channel names or choose `None`')

    # Compute bipolar EOG channel
    anode = channels[0]
    cathode = channels[1]
    print(f'Adding bipolar channel {new_name} ({anode} - {cathode})')
    raw = set_bipolar_reference(
        raw, anode, cathode, new_name, drop_refs=False, verbose=False)
    raw = raw.set_channel_types({new_name: 'eog'})

    return raw


def apply_montage(raw, montage):
    """Reads channel locations from custom file or standard montage."""

    # Load custom montage from file
    if path.isfile(montage):
        print(f'Loading custom montage from {montage}')
        digmontage = read_custom_montage(montage)

    # Or load standard montage
    else:
        print(f'Loading standard montage {montage}')
        digmontage = make_standard_montage(montage)

    # Make sure that EOG channels are of the `eog` type
    eog_channels = ['HEOG', 'VEOG', 'IO1', 'IO2', 'Afp9', 'Afp10', 'Auge_u']
    for ch_name in eog_channels:
        if ch_name in raw.ch_names:
            raw.set_channel_types({ch_name: 'eog'})

    # Make sure that mastoid channels are of the `misc` type
    misc_channels = ['A1', 'A2', 'M1', 'M2']
    for ch_name in misc_channels:
        if ch_name in raw.ch_names:
            raw.set_channel_types({ch_name: 'misc'})

    # Drop EEG channels that are not in the montage
    raw_channels = set(raw.copy().pick_types(eeg=True).ch_names)
    montage_channels = set(digmontage.ch_names)
    drop_channels = list(raw_channels - montage_channels)
    if drop_channels != []:
        print(f'Removing channels that are not in the montage {drop_channels}')
        raw.drop_channels(drop_channels)

    # Apply montage
    raw.set_montage(digmontage)


def interpolate_bad_channels(raw, bad_channels=None, auto_bad_channels=None):
    """Interpolates any channels from the two lists."""

    # Combine lists of bad channels
    all_bad_channels = []
    if bad_channels is not None and bad_channels != 'auto':
        all_bad_channels += bad_channels
    if auto_bad_channels is not None:
        all_bad_channels += auto_bad_channels

    # Interpolate bad channels
    if all_bad_channels != []:
        raw.info['bads'] += all_bad_channels
        raw = raw.interpolate_bads()

    return raw, all_bad_channels


def correct_ica(raw, method='fastica', n_components=0.99, random_seed=1234):
    """Corrects ocular artifacts using ICA and automatic component removal."""

    # Run ICA on a copy of the data
    raw_filt_ica = raw.copy()
    raw_filt_ica.load_data().filter(l_freq=1, h_freq=None, verbose=False)
    ica = ICA(
        n_components, random_state=random_seed, method=method, max_iter='auto')
    ica.fit(raw_filt_ica)

    # Remove bad components from the raw data
    eog_indices, _ = ica.find_bads_eog(
        raw, ch_name=['HEOG', 'VEOG'], verbose=False)
    ica.exclude = eog_indices
    raw = ica.apply(raw)

    return raw, ica


def correct_besa(raw, besa_file):
    """Corrects ocular artifacts using a pre-computed MSEC (BESA) matrix."""

    # Read BESA matrix
    print(f'Doing ocular correction with MSEC (BESA)')
    besa_matrix = pd.read_csv(besa_file, delimiter='\t', index_col=0)

    # Get EEG channel labels that are present in the data
    eeg_channels = raw.copy().pick_types(eeg=True).ch_names

    # Convert EEG channel labels to uppercase
    eeg_upper = pd.Series(eeg_channels).str.upper().values

    # Also convert BESA matrix labels to uppercase
    besa_matrix.index = besa_matrix.index.str.upper()
    besa_matrix.columns = besa_matrix.columns.str.upper()

    # Match so that the BESA matrix only contains channels that are in the data
    row_channels = [ch for ch in besa_matrix.index if ch in eeg_upper]
    col_channels = [ch for ch in besa_matrix.columns if ch in eeg_upper]
    besa_matrix = besa_matrix.reindex(index=row_channels, columns=col_channels)

    # Apply BESA matrix to the data
    eeg_data, _ = raw[eeg_channels]
    eeg_data = besa_matrix.values.dot(eeg_data)
    raw[eeg_channels] = eeg_data

    return raw



import json
import re
from glob import glob
from os import makedirs, path
from platform import python_version

import pandas as pd
from mne import Evoked
from mne import __version__ as mne_version
from mne import write_evokeds
from mne.channels.layout import _find_topomap_coords
from mne.io import concatenate_raws, read_raw_brainvision
from mne.time_frequency import AverageTFR, write_tfrs
from numpy import __version__ as numpy_version
from pandas import __version__ as pandas_version
from sklearn import __version__ as sk_version

from ._version import version as pipeline_version


def read_raw(vhdr_file_or_files):
    """Reads one or more raw EEG datasets from the same participant."""

    # Read raw datasets and combine if a list was provided
    if isinstance(vhdr_file_or_files, list):
        vhdr_files = vhdr_file_or_files
        print(f'\n=== Reading and combining raw data from {vhdr_files} ===')
        raw_list = [read_raw_brainvision(f, preload=True) for f in vhdr_files]
        raw = concatenate_raws(raw_list)
        participant_id = get_participant_id(vhdr_files)

    # Read raw dataset if only a single one was provided
    else:
        vhdr_file = vhdr_file_or_files
        print(f'\n=== Reading raw data from {vhdr_file} ===')
        raw = read_raw_brainvision(vhdr_file, preload=True)
        participant_id = get_participant_id(vhdr_file)

    return raw, participant_id


def get_participant_id(vhdr_file_or_files):
    """Extracts the basename of an input file to use as participant ID."""

    # Extract participant ID from raw data file name(s)
    if isinstance(vhdr_file_or_files, list):
        vhdr_files = vhdr_file_or_files
        participant_id = [path.basename(f).split('.')[0] for f in vhdr_files]
        participant_id = '_'.join(participant_id)
    else:
        vhdr_file = vhdr_file_or_files
        participant_id = path.basename(vhdr_file).split('.')[0]

    return participant_id


def files_from_dir(dir_path, extensions, natsort_files=True):
    """Retrieves files matching pattern(s) from a given parent directory."""

    # Find all files with one of the right extensions
    assert path.isdir(dir_path), f'Didn\'t find directory `{dir_path}`!'
    files = []
    for extension in extensions:
        files += glob(f'{dir_path}/*.{extension}')

    # Sort naturally because some files might not have leading zeros
    if natsort_files:
        files = sorted(files, key=natsort)

    return files


def natsort(s):
    """Natural-sort list of strings to handle (non-)leading zeros."""

    return [int(t) if t.isdigit() else t.lower() for t in re.split('(\d+)', s)]


def convert_participant_input(input, participant_ids):
    """Converts different inputs (e.g., dict) into a per-participant list."""

    # If it's a dict, convert to list
    if isinstance(input, dict):
        participant_dict = {id: None for id in participant_ids}
        for id, values in input.items():
            assert id in participant_ids, \
                f'Participant ID {id} is not in vhdr_files'
            values = [values] if not isinstance(values, list) else values
            participant_dict[id] = values
        return list(participant_dict.values())

    # If it's a list of list, it must have the same length as participant_ids
    elif is_nested_list(input):
        assert len(input) == len(participant_ids), \
            'Input lists must have the same length'
        return input

    # Otherwise all participants get the same values
    else:
        return [input] * len(participant_ids)


def is_nested_list(input):
    """Checks if a list is nested, i.e., contains at least one other list."""

    # Check if there is any list in the list
    if isinstance(input, list):
        return any(isinstance(elem, list) for elem in input)
    else:
        return False


def save_clean(raw, output_dir, participant_id=''):
    """Saves cleaned (continuous) EEG data in `.fif` format."""

    # Re-format participant ID for filename
    participant_id_ = '' if participant_id == '' else f'{participant_id}_'
    suffix = 'cleaned_eeg'

    # Create output folder and save
    makedirs(output_dir, exist_ok=True)
    fname = f'{output_dir}/{participant_id_}{suffix}.fif'
    raw.save(fname, overwrite=True)


def save_df(df, output_dir, participant_id='', suffix=''):
    """Saves pd.DataFrame in `.csv` format."""

    # Create output folder
    makedirs(output_dir, exist_ok=True)

    # Re-format participant ID and suffix for filename
    participant_id_ = '' if participant_id == '' else f'{participant_id}_'
    suffix = '' if suffix == '' else suffix

    # Save DataFrame
    fname = f'{output_dir}/{participant_id_}{suffix}.csv'
    df.to_csv(
        fname, na_rep='NA', float_format='%.4f', index=False)


def save_epochs(epochs, output_dir, participant_id='', to_df=True):
    """Saves mne.Epochs with metadata in `.fif` and/or `.csv` format."""

    # Create output folder
    makedirs(output_dir, exist_ok=True)

    # Re-format participant ID for filename
    participant_id_ = '' if participant_id == '' else f'{participant_id}_'
    suffix = 'epo'

    # Convert to DataFrame
    if to_df is True or to_df == 'both':
        scalings = {'eeg': 1e6, 'misc': 1e6}
        epochs_df = epochs.to_data_frame(scalings=scalings, time_format=None)
        epochs_df = epochs_df.rename(columns={'condition': 'event_id'})

        # Add metadata from log file
        metadata_df = epochs.metadata.copy()
        metadata_df = metadata_df.drop([col for col in metadata_df.columns
                                        if col in epochs_df.columns], axis=1)
        n_samples = len(epochs.times)
        metadata_df = metadata_df.loc[metadata_df.index.repeat(n_samples)]
        metadata_df = metadata_df.reset_index(drop=True)
        epochs_df = pd.concat([metadata_df, epochs_df], axis=1)

        # Save DataFrame
        save_df(epochs_df, output_dir, participant_id, suffix)

    # Save as MNE object
    if to_df is False or to_df == 'both':
        fname = f'{output_dir}/{participant_id_}{suffix}.fif'
        epochs.save(fname, overwrite=True)


def save_evokeds(
        evokeds, evokeds_df, output_dir, participant_id='', to_df=True):
    """Saves a list of mne.Evokeds in `.fif` and/or `.csv` format."""

    # Re-format participant ID for filename
    participant_id_ = '' if participant_id == '' else f'{participant_id}_'
    suffix = 'ave'

    # Create output directory
    makedirs(output_dir, exist_ok=True)

    # Save evokeds as DataFrame
    if to_df is True or to_df == 'both':
        save_df(evokeds_df, output_dir, participant_id, suffix)

    # Save evokeds as MNE object
    if to_df is False or to_df == 'both':

        # Save evokeds for ERPs
        if isinstance(evokeds[0], Evoked):
            fname = f'{output_dir}/{participant_id_}{suffix}.fif'
            write_evokeds(fname, evokeds, overwrite=True, verbose=False)

        # Save vokeds for TFR
        elif isinstance(evokeds[0], AverageTFR):
            fname = f'{output_dir}/{participant_id_}{suffix}.h5'
            write_tfrs(fname, evokeds, overwrite=True, verbose=False)


def save_montage(epochs, output_dir):
    """Saves channel locations in `.csv` format."""

    # Create output directory
    makedirs(output_dir, exist_ok=True)

    # Get locations of EEG channels
    chs = epochs.copy().pick_types(eeg=True).info['chs']
    coords = [ch['loc'][0:3] for ch in chs]
    coords_df = pd.DataFrame(
        columns=['cart_x', 'cart_y', 'cart_z'], data=coords)

    # Add channel names
    ch_names = [ch['ch_name'] for ch in chs]
    coords_df.insert(loc=0, column='channel', value=ch_names)

    # Add 2D flattened coordinates
    # Multiplied to mm scale (with head radius =~ 95 mm as in R-eegUtils)
    coords_df[['x', 'y']] = _find_topomap_coords(epochs.info, ch_names) * 947

    # Save
    save_df(coords_df, output_dir, suffix='channel_locations')


def save_config(config, output_dir):
    """Saves dict of pipeline config options in `.json` format."""

    # Create output directory
    makedirs(output_dir, exist_ok=True)

    # Save
    fname = f'{output_dir}/config.json'
    with open(fname, 'w') as f:
        json.dump(config, f, indent=4)


def save_report(report, output_dir, participant_id):
    """Saves HTML report."""

    # Create output directory
    makedirs(output_dir, exist_ok=True)

    # Save
    fname = f'{output_dir}/{participant_id}_report.html'
    print(f'Saving HTML report to {fname}')
    _ = report.save(fname, open_browser=False, overwrite=True)


def package_versions():
    """Returns pipeline version and important dependency package versions."""

    return ({'python': python_version(),
             'pipeline': pipeline_version,
             'mne': mne_version,
             'numpy': numpy_version,
             'pandas': pandas_version,
             'scikit-learn': sk_version})



import chardet
import numpy as np
import pandas as pd
from mne import combine_evoked, pick_channels, set_log_level
from mne.channels import combine_channels
from scipy.stats import zscore


def triggers_to_event_id(triggers):
    """Convert list or dict of triggers to MNE-style event_id"""

    # Convert list to dict with triggers as condition names
    if isinstance(triggers, list):
        triggers = {str(trigger): trigger for trigger in triggers}
    else:
        assert isinstance(triggers, dict), \
            '`triggers` must be either list or dict'

    # Make sure that trigger values are integers (R would pass them as floats)
    event_id = {key: int(value) for key, value in triggers.items()}

    return event_id


def read_log(log_file, skip_log_rows=None, skip_log_conditions=None):
    """Reads the behavioral log file with information about each EEG trial."""

    # Check if data are already in a DataFrame
    if isinstance(log_file, pd.DataFrame):
        log = log_file
    else:

        # Detect file encoding
        with open(log_file, 'rb') as f:
            data = f.read()
        chardet_res = chardet.detect(data)
        encoding = chardet_res['encoding']

        # Read into DataFrame
        if '.csv' in log_file:
            log = pd.read_csv(log_file, encoding=encoding)
        else:
            log = pd.read_csv(log_file, delimiter='\t', encoding=encoding)

    # Remove rows via indices (e.g., if the EEG was paused accidently)
    if skip_log_rows is not None:
        log = log.drop(skip_log_rows)

    # Remove rows via conditions (e.g., for filler stimuli without triggers)
    if skip_log_conditions is not None:
        assert isinstance(skip_log_conditions, dict), \
            '"skip_log_conditions" must be a dict ({column: [conditions]})'
        for col, values in skip_log_conditions.items():
            if not isinstance(values, list):
                log = log[log[col] != values]
            else:
                log = log[~log[col].isin(values)]

    return log


def match_log_to_epochs(epochs, log, triggers_column, depth=10):
    """Auto-detects missing EEG trials and removes them from the log file."""

    # Make sure that the requested column is available in the log file
    assert triggers_column in log.columns, \
        f'Column \'{triggers_column}\' is not in the log file'

    # Read lists of triggers from log file and EEG epochs
    events_log = log[triggers_column].tolist()
    events_epochs = list(epochs.events[:, 2])

    # Check for each row in the log file
    previous_repaired = False
    for ix in range(len(events_log)):

        # Add `nan` in case trials are missing at the end of the EEG...
        if len(events_epochs) <= ix:
            print(f'Log file (row index {ix}): Found missing EEG epoch')
            events_epochs.insert(ix, np.nan)

        # ... or if the log and EEG trigers don't match up
        elif events_log[ix] != events_epochs[ix]:
            print(f'Log file (row index {ix}): Found missing EEG epoch')
            events_epochs.insert(ix, np.nan)
            previous_repaired = True

        # If they do match up, we check that the next trials do match as well
        elif previous_repaired:
            if events_log[ix:ix + depth] != events_epochs[ix:ix + depth]:
                print(f'Log file (row index {ix}): Assuming missing EEG epoch')
                events_epochs.insert(ix, np.nan)
            else:
                previous_repaired = False

    # Remove trials with missing EEG epochs from the log file
    missing_ixs = np.where(np.isnan(events_epochs))[0].tolist()
    print(f'Dropping rows from the log file data: {missing_ixs}')
    log = log.reset_index(drop=True)
    log = log.drop(index=missing_ixs)

    return log, missing_ixs


def get_bad_epochs(epochs, reject_peak_to_peak=None):
    """Detects bad epochs based on peak-to-peak amplitude."""

    # Convert thresholds to volts
    if reject_peak_to_peak is not None:
        reject_peak_to_peak = {'eeg': reject_peak_to_peak * 1e-6}

    # Reject on a copy of the data
    epochs_rej = epochs.copy().drop_bad(reject_peak_to_peak)

    # Get indices of bad epochs from the rejection log
    all_ixs = [elem for elem in epochs_rej.drop_log if elem != ('IGNORED',)]
    bad_ixs = [ix for ix, elem in enumerate(all_ixs) if elem != ()]

    return bad_ixs


def get_bad_channels(epochs, threshold=3., by_event_type=True):
    """Automatically detects bad channels using their average standard error"""

    # Compute standard error for each condition seperately, then average...
    if by_event_type:
        ses = epochs.standard_error(by_event_type=True)
        ses = combine_evoked(ses, weights='nave')

    # ... or directly compute standard error across all epochs
    else:
        ses = epochs.standard_error()

    # Average across time points for each channel
    ses = ses.data.mean(axis=1)

    # Convert to z scores
    zs = zscore(ses)

    # Look up bad channel labels
    ixs = np.where(zs > threshold)[0]
    bad_channels = [epochs.ch_names[ix] for ix in ixs]
    if bad_channels != []:
        print(f'Automatically detected bad channels {bad_channels} with '
              f'z_SE > {threshold}')
    else:
        print(f'Didn\'t detect any bad channels with z_SE > {threshold}')

    return bad_channels


def compute_single_trials(epochs, components, bad_ixs=None):
    """Computes single trial mean amplitudes a dict of multiple components."""

    # Check that values in the dict are lists
    if not isinstance(components['name'], list):
        components = {key: [value] for key, value in components.items()}

    # Loop over components
    components_df = pd.DataFrame(components)
    for _, component in components_df.iterrows():

        # Compute single trial mean ERP amplitudes
        compute_component(
            epochs, component['name'], component['tmin'],
            component['tmax'], component['roi'], bad_ixs)

    return epochs.metadata


def compute_component(epochs, name, tmin, tmax, roi, bad_ixs=None):
    """Computes single trial mean amplitudes for single component."""

    # Check that requested region of interest channels are present in the data
    for ch in roi:
        assert ch in epochs.ch_names, f'ROI channel \'{ch}\' not in the data'

    # Create virtual channel for the average in the region of interest
    print(f'Computing single trial ERP amplitudes for \'{name}\'')
    set_log_level('ERROR')
    roi_dict = {name: pick_channels(epochs.ch_names, roi)}
    epochs_roi = combine_channels(epochs, roi_dict)
    epochs.add_channels([epochs_roi], force_update_info=True)
    epochs.set_channel_types({name: 'misc'})

    # Compute mean amplitudes by averaging across the relevant time window
    epochs_roi.crop(tmin, tmax)
    df = epochs_roi.to_data_frame()
    mean_amp = df.groupby('epoch')[name].mean()

    # Set ERPs for bad epochs to NaN
    if bad_ixs is not None:
        if isinstance(bad_ixs, int):
            bad_ixs = [bad_ixs]
        mean_amp[bad_ixs] = np.nan

    # Add as a new column to the original metadata
    epochs.metadata.reset_index(drop=True, inplace=True)
    epochs.metadata = pd.concat([epochs.metadata, mean_amp], axis=1)
    set_log_level('INFO')



from functools import partial

import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from .averaging import compute_grands, compute_grands_df
from .io import (convert_participant_input, files_from_dir, get_participant_id,
                 package_versions, save_config, save_df, save_evokeds)
from .participant import participant_pipeline
from .perm import compute_perm, compute_perm_tfr


def group_pipeline(
    vhdr_files,
    log_files,
    output_dir,
    clean_dir=None,
    epochs_dir=None,
    report_dir=None,
    to_df=True,
    downsample_sfreq=None,
    veog_channels='auto',
    heog_channels='auto',
    montage='easycap-M1',
    bad_channels=None,
    besa_files=None,
    ica_method=None,
    ica_n_components=0.99,
    highpass_freq=0.1,
    lowpass_freq=40.0,
    triggers=None,
    triggers_column=None,
    epochs_tmin=-0.5,
    epochs_tmax=1.5,
    baseline=(-0.2, 0.0),
    skip_log_rows=None,
    skip_log_conditions=None,
    reject_peak_to_peak=200.0,
    components={'name': [], 'tmin': [], 'tmax': [], 'roi': []},
    average_by=None,
    perform_tfr=False,
    tfr_subtract_evoked=False,
    tfr_freqs=np.linspace(4.0, 40.0, num=37),
    tfr_cycles=np.linspace(2.0, 20.0, num=37),
    tfr_mode='percent',
    tfr_baseline=(-0.45, -0.05),
    tfr_components={
        'name': [], 'tmin': [], 'tmax': [], 'fmin': [], 'fmax': [], 'roi': []},
    perm_contrasts=[],
    perm_tmin=0.0,
    perm_tmax=1.0,
    perm_channels=None,
    perm_fmin=None,
    perm_fmax=None,
    n_jobs=1
):
    """Processes EEG data for all participants of an experiment.

    For each participant, the raw data is read and cleaned using standard steps
    (downsampling, bad channel interpolation, ocular correction, frequency
    domain filtering). Epochs are created around the `triggers`. Bad epochs are
    removed based on peak-to-peak amplitude. Single trial mean ERP amplitudes
    for ERP `components` of interest are computed and matched to the single
    trial behavioral data from the `log_files`.

    Optionally, this last step is repeated on a time-frequency representation
    (TFR) of the data obtained via Morlet wavelet convolution.

    The result is a single trial data frame which can be used for fitting
    linear mixed-effects models on the mean ERP amplitudes (and power).

    Additionally, by-participant condition averages (`evokeds`) for the ERPs
    (and power) are computed to facilitate plotting. Optionally, these can also
    be tested for condition differences in an exploratory fashion using
    cluster-based permutation tests.

    For details about the pipeline, see Frömer et al. (2018)[1].

    Parameters & returns
    --------------------
    See the README[2] in the GitHub repository for the pipeline.

    Notes
    -----
    [1] https://doi.org/10.3389/fnins.2018.00048
    [2] https://github.com/alexenge/hu-neuro-pipeline/blob/dev/README.md
    """

    # Convert input types
    tfr_freqs = list(tfr_freqs)
    tfr_cycles = list(tfr_cycles)

    # Backup input arguments for re-use
    config = locals().copy()

    # Create partial function with arguments shared across participants
    partial_pipeline = partial(
        participant_pipeline,
        skip_log_conditions=skip_log_conditions,
        downsample_sfreq=downsample_sfreq,
        veog_channels=veog_channels,
        heog_channels=heog_channels,
        montage=montage,
        ica_method=ica_method,
        ica_n_components=ica_n_components,
        highpass_freq=highpass_freq,
        lowpass_freq=lowpass_freq,
        triggers=triggers,
        triggers_column=triggers_column,
        epochs_tmin=epochs_tmin,
        epochs_tmax=epochs_tmax,
        baseline=baseline,
        reject_peak_to_peak=reject_peak_to_peak,
        components=components,
        average_by=average_by,
        perform_tfr=perform_tfr,
        tfr_subtract_evoked=tfr_subtract_evoked,
        tfr_freqs=tfr_freqs,
        tfr_cycles=tfr_cycles,
        tfr_mode=tfr_mode,
        tfr_baseline=tfr_baseline,
        tfr_components=tfr_components,
        clean_dir=clean_dir,
        epochs_dir=epochs_dir,
        chanlocs_dir=output_dir,
        report_dir=report_dir,
        to_df=to_df)

    # Get input file paths if directories were provided
    if isinstance(vhdr_files, str):
        vhdr_files = files_from_dir(vhdr_files, extensions=['vhdr'])
    if isinstance(log_files, str):
        log_files = files_from_dir(log_files, extensions=['csv', 'tsv', 'txt'])
    assert len(log_files) == len(vhdr_files), \
        f'Number of `log_files` ({len(log_files)}) does not match ' + \
        f'number of `vhdr_files` ({len(vhdr_files)})'

    # Get input BESA matrix files if necessary
    if isinstance(besa_files, str):
        besa_files = files_from_dir(besa_files, extensions=['matrix'])
    elif besa_files is None:
        besa_files = [None] * len(vhdr_files)
    assert len(besa_files) == len(vhdr_files), \
        f'Number of `besa_files` ({len(besa_files)}) does not match ' + \
        f'number of `vhdr_files` ({len(vhdr_files)})'

    # Extract participant IDs from filenames
    participant_ids = [get_participant_id(f) for f in vhdr_files]

    # Construct lists of bad_channels and skip_log_rows per participant
    bad_channels = convert_participant_input(bad_channels, participant_ids)
    skip_log_rows = convert_participant_input(skip_log_rows, participant_ids)

    # Combine participant-specific inputs
    participant_args = zip(vhdr_files, log_files, besa_files,
                           bad_channels, skip_log_rows)

    # Do processing in parallel
    n_jobs = int(n_jobs)
    res = Parallel(n_jobs)(
        delayed(partial_pipeline)(*args) for args in participant_args)

    # Sort outputs into seperate lists
    print(f'\n\n=== Processing group level ===')
    trials, evokeds, evokeds_dfs, configs = list(map(list, zip(*res)))[0:4]

    # Combine trials and save
    trials = pd.concat(trials, ignore_index=True)
    save_df(trials, output_dir, suffix='trials')

    # Combine evokeds_dfs and save
    evokeds_df = pd.concat(evokeds_dfs, ignore_index=True)
    save_df(evokeds_df, output_dir, suffix='ave')

    # Compute grand averaged ERPs and save
    grands = compute_grands(evokeds)
    grands_df = compute_grands_df(evokeds_df)
    save_evokeds(
        grands, grands_df, output_dir, participant_id='grand', to_df=to_df)

    # Update config with participant-specific inputs...
    config['vhdr_files'] = vhdr_files
    config['bad_channels'] = bad_channels
    config['besa_files'] = besa_files
    config['skip_log_rows'] = skip_log_rows

    # ... and outputs that might have been created along the way
    config['log_files'] = []
    if triggers_column is not None:
        config['auto_missing_epochs'] = {}
    if reject_peak_to_peak is not None:
        config['auto_rejected_epochs'] = {}
    if ica_method is not None and ica_n_components < 1.0:
        config['auto_ica_n_components'] = {}
    for pid, pconfig in zip(participant_ids, configs):
        config['log_files'].append(pconfig['log_file'])
        if triggers_column is not None:
            config['auto_missing_epochs'][pid] = \
                pconfig['auto_missing_epochs']
        if reject_peak_to_peak is not None:
            config['auto_rejected_epochs'][pid] = \
                pconfig['auto_rejected_epochs']
        if pconfig['bad_channels'] == 'auto':
            config.setdefault('auto_bad_channels', {}).update(
                {pid: pconfig['auto_bad_channels']})
        if pconfig['ica_method'] is not None:
            if ica_n_components < 1.0:
                config['auto_ica_n_components'][pid] = \
                    pconfig['auto_ica_n_components']
            config.setdefault('auto_ica_bad_components', {}).update(
                {pid: pconfig['auto_ica_bad_components']})

    # Save config
    config['package_versions'] = package_versions()
    save_config(config, output_dir)

    # Define standard returns
    returns = [trials, evokeds_df, config]

    # Cluster based permutation tests for ERPs
    if perm_contrasts != []:
        cluster_df = compute_perm(evokeds, perm_contrasts, perm_tmin,
                                  perm_tmax, perm_channels, n_jobs)
        save_df(cluster_df, output_dir, suffix='clusters')
        returns.append(cluster_df)

    # Combine time-frequency results
    if perform_tfr:

        # Sort outputs into seperate lists
        tfr_evokeds, tfr_evokeds_dfs = list(map(list, zip(*res)))[4:6]

        # Combine evokeds_df for power and save
        tfr_evokeds_df = pd.concat(tfr_evokeds_dfs, ignore_index=True)
        save_df(tfr_evokeds_df, output_dir,
                suffix='tfr_ave')
        returns.append(tfr_evokeds_df)

        # Compute grand averaged power and save
        tfr_grands = compute_grands(tfr_evokeds)
        tfr_grands_df = compute_grands_df(tfr_evokeds_df)
        save_evokeds(tfr_grands, tfr_grands_df, output_dir,
                     participant_id='tfr_grand', to_df=to_df)

        # Cluster based permutation tests for TFR
        if perm_contrasts != []:
            tfr_cluster_df = compute_perm_tfr(
                tfr_evokeds, perm_contrasts, perm_tmin, perm_tmax,
                perm_channels, perm_fmin, perm_fmax, n_jobs)
            save_df(tfr_cluster_df, output_dir, suffix='tfr_clusters')
            returns.append(tfr_cluster_df)

    return returns



import numpy as np
import pandas as pd
from mne import concatenate_epochs, set_log_level


def subtract_evoked(epochs, average_by=None, evokeds=None):
    """Subtracts evoked activity (across or by conditions) from epochs."""

    # If no columns were requested, subtract evoked activity across conditions
    set_log_level('ERROR')
    if average_by is None:
        print('Subtracting evoked activity')
        epochs = epochs.subtract_evoked()

    # Otherwise subtract seperately for all (combinations of) conditions
    else:
        print('Subtracting evoked activity per condition in `average_by`')
        epochs = subtract_evoked_conditions(epochs, average_by, evokeds)
    set_log_level('INFO')

    return epochs


def subtract_evoked_conditions(epochs, average_by, evokeds):
    """Subtracts evoked activity (separately by conditions) from epochs."""

    # Loop over epochs (painfully slow)
    epochs_subtracted = []
    for ix, _ in enumerate(epochs):
        for query, evoked in zip(average_by.values(), evokeds):
            if len(epochs[ix][query]) > 0:
                epoch_subtracted = epochs[ix].subtract_evoked(evoked)
                epochs_subtracted.append(epoch_subtracted)

    return concatenate_epochs(epochs_subtracted)


def compute_single_trials_tfr(epochs, components, bad_ixs=None):
    """Computes single trial power for a dict of multiple components."""

    # Check that values in the dict are lists
    if not isinstance(components['name'], list):
        components = {key: [value] for key, value in components.items()}

    # Loop over components
    components_df = pd.DataFrame(components)
    for _, component in components_df.iterrows():

        # Comput single trial power
        compute_component_tfr(
            epochs, component['name'], component['tmin'],
            component['tmax'], component['fmin'], component['fmax'],
            component['roi'], bad_ixs)

    return epochs.metadata


def compute_component_tfr(
        epochs, name, tmin, tmax, fmin, fmax, roi, bad_ixs=None):
    """Computes single trial power for a single component."""

    # Check that requested region of interest channels are present in the data
    for ch in roi:
        assert ch in epochs.ch_names, f'ROI channel \'{ch}\' not in the data'

    # Select region, time window, and frequencies of interest
    print(f'Computing single trial power amplitudes for \'{name}\'')
    epochs_oi = epochs.copy().pick_channels(roi).crop(tmin, tmax, fmin, fmax)

    # Compute mean power per trial
    mean_power = epochs_oi.data.mean(axis=(1, 2, 3))

    # Set power for bad epochs to NaN
    if bad_ixs is not None:
        if isinstance(bad_ixs, int):
            bad_ixs = [bad_ixs]
        mean_power[bad_ixs] = np.nan

    # Add as a new column to the original metadata
    epochs.metadata[name] = mean_power



import json
import re
import urllib.request

import pooch

# URL of the UCAP repo files on OSF
base_url = 'https://files.de-1.osf.io/v1/resources/hdxvb/providers/osfstorage'

# Local cache to download the files to
local_cache = 'hu-neuro-pipeline'


def construct_fetcher(remote_dir, local_dir):
    """Constructs a pooch fetcher for getting UCAP remote files locally."""

    # Parse remote files on OSF
    urls = {}
    hashes = {}
    with urllib.request.urlopen(f'{base_url}/{remote_dir}') as url:
        files = json.loads(url.read().decode())['data']
        files_attributes = [file['attributes'] for file in files]

        # Natural-sort files by name
        # Necessary because log file names are without leading zeros
        natsort = lambda s: [int(t) if t.isdigit() else t.lower()
                             for t in re.split('(\d+)', s)]
        files_attributes_sorted = sorted(
            files_attributes, key=lambda d: natsort(d['name']))

        # Extract file URLs and hashsums
        for attributes in files_attributes_sorted:
            local_path = local_dir + attributes['name']
            urls[local_path] = base_url + attributes['path']
            hashes[local_path] = 'md5:' + attributes['extra']['hashes']['md5']

    # Construct fetcher for getting UCAP data
    fetcher = pooch.create(
        path=pooch.os_cache(local_cache),
        base_url=base_url,
        env='PIPELINE_DATA_DIR',
        registry=hashes,
        urls=urls,
    )

    return fetcher


def get_paths(n_participants=40):
    """Downloads sample data from the UCAP study and returns the paths."""

    # Prepare dict of file paths for return
    paths = {}

    # Get raw EEG files - note that there are 3 files per participant
    eeg_fetcher = construct_fetcher('59cf07fa6c613b02958f3364/', 'raw/')
    n_files = int(n_participants) * 3
    eeg_paths = list(eeg_fetcher.registry.keys())[:n_files]
    eeg_paths = [eeg_fetcher.fetch(path) for path in eeg_paths]
    vhdr_paths = [path for path in eeg_paths if path.endswith('.vhdr')]
    paths['vhdr_files'] = vhdr_paths

    # Extract participant IDs
    participant_ids = [
        path.split('/')[-1].replace('.vhdr', '') for path in vhdr_paths]

    # Get corresponding behavioral log files
    log_fetcher = construct_fetcher('59cf12259ad5a102cc5c4b93/', 'log/')
    log_paths = [f'log/{int(p_id)}_test.txt' for p_id in participant_ids]
    log_paths = [log_fetcher.fetch(path) for path in log_paths]
    paths['log_files'] = log_paths

    # Get corresponding BESA/MSEC cali files
    cali_fetcher = construct_fetcher('59cf089e6c613b02968f5724/', 'cali/')
    cali_paths = [f'cali/{p_id}_cali.matrix' for p_id in participant_ids]
    cali_paths = [cali_fetcher.fetch(path) for path in cali_paths]
    paths['besa_files'] = cali_paths

    return paths



import warnings

import numpy as np
import pandas as pd
from mne import Epochs, Evoked, grand_average
from mne.time_frequency import AverageTFR, EpochsTFR


def compute_evokeds(epochs, average_by=None, bad_ixs=[], participant_id=None):
    """Computes condition averages (evokeds) based on triggers or columns."""

    # Average by triggers in case no log file columns were provided
    if average_by is None:
        all_evokeds, all_evokeds_df = compute_evokeds_triggers(
            epochs, bad_ixs, participant_id)
    elif isinstance(average_by, dict):
        all_evokeds, all_evokeds_df = compute_evokeds_queries(
            epochs, average_by, bad_ixs, participant_id)
    else:
        warnings.warn(
            'Passing a list of column names to `average_by` will ' +
            'be deprecated in a future version of the pipeline. ' +
            'Please use a dict of  labels and log file queries ' +
            'instead (see https://github.com/alexenge/hu-neuro-pipeline/blob/main/docs/inputs.md#average_by-recommended-default-none)')
        all_evokeds, all_evokeds_df = compute_evokeds_cols(
            epochs, average_by, bad_ixs, participant_id)

    return all_evokeds, all_evokeds_df


def compute_evokeds_triggers(epochs, bad_ixs=[], participant_id=None):
    """Computes condition averages (evokeds) based on triggers."""

    # Get indices of good epochs
    good_ixs = [ix for ix in range(len(epochs)) if ix not in bad_ixs]

    # Prepare emtpy lists
    all_evokeds = []
    all_evokeds_dfs = []

    # Compute evokeds
    epochs_good = epochs.copy()[good_ixs]
    evokeds = average_by_events(epochs_good)
    all_evokeds = all_evokeds + evokeds

    # Convert to DataFrame
    evokeds_df = create_evokeds_df(evokeds, participant_id=participant_id)
    all_evokeds_dfs.append(evokeds_df)

    # Combine DataFrames
    all_evokeds_df = pd.concat(all_evokeds_dfs, ignore_index=True)

    return all_evokeds, all_evokeds_df


def compute_evokeds_queries(epochs, queries, bad_ixs=[], participant_id=None):
    """Computes condition averages (evokeds) based on log file queries."""

    # Get indices of good epochs
    good_ixs = [ix for ix in range(len(epochs)) if ix not in bad_ixs]

    # Reset index so that trials start at 0
    epochs.metadata.reset_index(drop=True, inplace=True)

    # Create evokeds for each query
    evokeds = []
    evoked_dfs = []
    for label, query in queries.items():

        # Compute evokeds for trials that match the current query
        evoked = compute_evoked_query(epochs[good_ixs], query, label)
        evokeds.append(evoked)

        # Convert to data frame
        extra_cols = \
            {'participant_id': participant_id, 'label': label, 'query': query}
        evoked_df = evoked_to_df(evoked, extra_cols)
        evoked_dfs.append(evoked_df)

    # Combine data frames
    evokeds_df = pd.concat(evoked_dfs, ignore_index=True)

    return evokeds, evokeds_df


def compute_evoked_query(epochs, query, label):
    """Computes one condition average (evoked) based on a log file query."""

    # Compute evokeds based on ERP or TFR epochs
    if isinstance(epochs, EpochsTFR):
        evoked = epochs[query].average()
    else:  # `EpochsTFR.average()` has no `picks` argument
        evoked = epochs[query].average(picks=['eeg', 'misc'])
    evoked.comment = label

    return evoked


def evoked_to_df(evoked, extra_cols={}):
    """Converts MNE's Evoked or AverageTFR to a pandas data frame."""

    # Convert to data frame
    if isinstance(evoked, AverageTFR):
        evoked_df = evoked.to_data_frame()
    else:  # `AverageTFR.to_data_frame()` has no `scalings` argument
        evoked_df = \
            evoked.to_data_frame(scalings={'eeg': 1e6, 'misc': 1e6})

    # Optionally add extra columns
    for column, value in reversed(extra_cols.items()):
        evoked_df.insert(0, column, value)

    return evoked_df


def compute_evokeds_cols(
        epochs, average_by=None, bad_ixs=[], participant_id=None):
    """Computes condition averages (evokeds) based on log file columns."""

    # Make sure that provided values are stored in a list
    if isinstance(average_by, str):
        average_by = [average_by]

    # Get indices of good epochs
    good_ixs = [ix for ix in range(len(epochs)) if ix not in bad_ixs]

    # Prepare emtpy lists
    all_evokeds = []
    all_evokeds_dfs = []

    # Iterate over the provided main effects and interactions
    for cols in average_by:

        # Parse interaction effects into a list
        cols = cols.split('/')

        # Compute evokeds
        epochs_update = update_events(epochs, cols)[good_ixs]
        evokeds = average_by_events(epochs_update)
        all_evokeds = all_evokeds + evokeds

        # Convert to DataFrame
        trials = epochs_update.metadata
        evokeds_df = create_evokeds_df(
            evokeds, cols, trials, participant_id)

        # Append info about averaging
        value = '/'.join(cols)
        evokeds_df.insert(loc=1, column='average_by', value=value)
        all_evokeds_dfs.append(evokeds_df)

    # Combine DataFrames
    all_evokeds_df = pd.concat(all_evokeds_dfs, ignore_index=True)

    # Move condition columns back to the front
    # They might have been moved to the end while concatenating
    if average_by is not None:
        time_ix = all_evokeds_df.columns.get_loc('time')
        for cols in reversed(average_by):
            if not '/' in cols:
                all_evokeds_df.insert(
                    time_ix - 1, column=cols, value=all_evokeds_df.pop(cols))

                # Convert NaNs to empty strings so that R can represent them
                all_evokeds_df[cols] = all_evokeds_df[cols].fillna('')

    return all_evokeds, all_evokeds_df


def average_by_events(epochs, method='mean'):
    """Create a list of evokeds from epochs, one per event type."""

    # Pick channel types for ERPs
    # The `average` method for `EpochsTFR` doesn't support `picks`
    picks_dict = {'picks': ['eeg', 'misc']} \
        if isinstance(epochs, Epochs) else {}

    # Loop over event types and average
    # TODO: Use MNE built-in argument `by_event_type` once it's in `EpochsTFR`
    evokeds = []
    for event_type in epochs.event_id.keys():
        evoked = epochs[event_type].average(**picks_dict, method=method)
        evoked.comment = event_type
        evokeds.append(evoked)

    return evokeds


def update_events(epochs, cols):
    """Updates the events/event_id structures using cols from the metadata."""

    # Generate event codes for the relevant columns
    cols_df = pd.DataFrame(epochs.metadata[cols])
    cols_df = cols_df.astype('str')
    ids = cols_df.agg('/'.join, axis=1)
    codes = ids.astype('category').cat.codes

    # Create copy of the data with the new event codes
    epochs_update = epochs.copy()
    epochs_update.events[:, 2] = codes
    epochs_update.event_id = dict(zip(ids, codes))

    return epochs_update


def create_evokeds_df(evokeds, cols=None, trials=None, participant_id=None):
    """Converts mne.Evoked into a pd.DataFrame with metadata."""

    # Convert ERP amplitudes from volts to microvolts
    # The `to_data_frame` method for `AverageTFR` doesn't support `scalings`
    scalings_dict = {'scalings': {'eeg': 1e6, 'misc': 1e6}} \
        if isinstance(evokeds[0], Evoked) else {}

    # Convert all evokeds to a single DataFrame
    evokeds_dfs = [evoked.to_data_frame(**scalings_dict, time_format=None)
                   for evoked in evokeds]
    evokeds_df = pd.concat(evokeds_dfs, ignore_index=True)

    # Optionally add columns from the metadata
    repeats = len(evokeds_df)
    if cols is not None:
        assert trials is not None, 'Must provide trials (metadata) with cols'
        cols_df = pd.DataFrame(trials[cols])
        cols_df = cols_df.astype('str')
        cols_df = cols_df.drop_duplicates()
        repeats = len(evokeds_df) / len(cols_df)
        cols_df = cols_df.loc[cols_df.index.repeat(repeats)]
        cols_df = cols_df.reset_index(drop=True)
        evokeds_df = pd.concat([cols_df, evokeds_df], axis=1)

    # Otherwise add comments from evokeds (assumed to contain event IDs)
    else:
        comments = [evoked.comment for evoked in evokeds]
        repeats = len(evokeds_df) / len(comments)
        comments = np.repeat(comments, repeats)
        evokeds_df.insert(loc=0, column='event_id', value=comments)

    # Optionally add participant_id
    if participant_id is not None:
        evokeds_df.insert(loc=0, column='participant_id', value=participant_id)

    return evokeds_df


def compute_grands(evokeds_per_participant):
    """Averages evokeds of all participants into grand averages."""

    # Average across participants for each condition
    evokeds_per_condition = list(map(list, zip(*evokeds_per_participant)))
    grands = [grand_average(x) for x in evokeds_per_condition]

    # Add meaningful comments
    comments = [x[0].comment for x in evokeds_per_condition]
    for grand, comment in zip(grands, comments):
        grand.comment = comment

    return grands


def compute_grands_df(evokeds_df):
    """Averages evoked DataFrames of all participants into grand averages."""

    # Get indices of columns to group by (conditions, times, frequencies)
    first_grouping_ix = 1  # Column 0 is participant_id (to average over)
    last_grouping_col = 'freq' if 'freq' in evokeds_df.columns else 'time'
    last_grouping_ix = evokeds_df.columns.get_loc(last_grouping_col)
    grouping_ixs = range(first_grouping_ix, last_grouping_ix + 1)

    # Average by grouping columns
    group_cols = list(evokeds_df.columns[grouping_ixs])
    grands_df = evokeds_df.groupby(group_cols, dropna=False).mean()

    # Convert conditions from index back to columns
    grands_df = grands_df.reset_index()

    return grands_df



"""Single trial EEG pipeline at the Abdel Rahman Lab for Neurocognitive
Psychology, Humboldt-Universität zu Berlin"""

# Import submodules
from . import datasets

# Make central functions available as top level imports
from .group import group_pipeline
from .participant import participant_pipeline

# Get current package version
try:
    from ._version import version as __version__
    from ._version import version_tuple
except ImportError:
    __version__ = "unknown version"
    version_tuple = (0, 0, "unknown version")



import os
import os.path as op

import setuptools


def package_tree(pkgroot):
    """Get the submodule list  (adapted from MNE/VisPy)."""

    # Get all submodules based on their `__init__.py`
    path = op.dirname(__file__)
    subdirs = [op.relpath(i[0], path).replace(op.sep, '.')
               for i in os.walk(op.join(path, pkgroot))
               if '__init__.py' in i[2]]

    return sorted(subdirs)


if __name__ == "__main__":

    # Paste README as long description
    with open('README.md', 'r', encoding='utf-8') as fh:
        long_description = fh.read()

    # Actual setup
    setuptools.setup(
        name='hu-neuro-pipeline',
        author='Alexander Enge',
        author_email='alexander.enge@hu-berlin.de',
        description='Single trial EEG pipeline at the Abdel Rahman Lab for '
                    'Neurocognitive Psychology, Humboldt-Universität zu Berlin',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/alexenge/hu-neuro-pipeline',
        project_urls={
            'Issue trackers': 'https://github.com/alexenge/hu-neuro-pipeline/issues',
        },
        classifiers=[
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
        ],
        packages=package_tree('pipeline'),
        install_requires=[
            'chardet',
            'joblib',
            'matplotlib',
            'mne>=0.24.0',
            'pandas!=1.4.0',
            'pooch>=1.5',
            'scikit-learn'
        ],
        python_requires='>=3.8',
    )



from platform import python_version
from statistics import mean, median

import mne


def boilerplate(config):
    """Auto-creates part of the methods section based on pipeline options."""

    # Prepare empty lists
    boilerplate = []
    references = []

    # Pipeline intro
    text = (
        'The continuous EEG from {n_participants} participants was processed '
        'offline using the single trial EEG pipeline proposed by Frömer et '
        'al. (2018). The pipeline was implemented with the packages MNE '
        '(Version {mne_version}; Gramfort et al., 2013) and STEP-MNE '
        '({step_mne_url}) for Python (Version {python_version}; Van Rossum & '
        'Drake, 2009).\n')
    text = text.format(
        n_participants=len(config['vhdr_files']),
        mne_version=mne.__version__,
        step_mne_url='https://github.com/alexenge/step-mne',  # TODO: Add version
        python_version=python_version()
    )
    boilerplate.append(text)
    references.append(
        'Frömer, R., Maier, M., & Abdel Rahman, R. (2018). Group-level EEG-'
        'processing pipeline for flexible single trial-based analyses '
        'including linear mixed models. Frontiers in Neuroscience, 12, 48. '
        'https://doi.org/10.3389/fnins.2018.00048')
    references.append(
        'Gramfort, A., Luessi, M., Larson, E., Engemann, D. A., Strohmeier, '
        'D., Brodbeck, C., Goj, R., Jas, M., Brooks, T., Parkkonen, L., & '
        'Hämäläinen, M. (2013). MEG and EEG data analysis with MNE-Python. '
        'Frontiers in Neuroscience, 7. '
        'https://doi.org/10.3389/fnins.2013.00267')
    references.append(
        'Van Rossum, G., & Drake, F. L. (2009). Python 3 reference manual. '
        'CreateSpace.')
    
    # Downsampling
    if config['downsample_sfreq'] is not None:
        text = (
            'The data from each participant were downsampled to '
            '{downsample_sfreq} Hz. '
        )
        text = text.format(downsample_sfreq = int(config['downsample_sfreq']))
        boilerplate.append(text)

    # Bad channels
    ns_bads = [len(l) for l in config['bad_channels']]
    if sum(ns_bads) > 0:
        if config['bad_channels'] == 'auto':
            text = (
                'An average of {mean_bads} EEG channels per participant '
                '(Mdn = {mdn_bads}, range {min_bads} to {max_bads}) were '
                'automatically flagged for bad data quality. Channels were '
                'flagged as bad if their inclusion would have led to the '
                'rejection of at least 5% of all available epochs for the '
                'participant, according to the artifact rejection threshold '
                'defined below. The signal for these channels was replaced by '
                'the signal of the neighboring channels using spherical '
                'spline interpolation (Perrin et al., 1989). '
            )
        else:
            text = (
                'An average of {mean_bads} EEG channels per participant '
                '(Mdn = {mdn_bads}, range {min_bads} to {max_bads}) were '
                'manually flagged for bad data quality. The signal for these '
                'channels was replaced by the signal of the neighboring '
                'channels using spherical spline interpolation (Perrin et '
                'al., 1989). '
            )
        text = text.format(
            mean_bads = '{:.1f}'.format(mean(ns_bads)),
            mdn_bads = median(ns_bads),
            min_bads = min(ns_bads),
            max_bads = max(ns_bads)
        )
        boilerplate.append(text)
        references.append(
            'Perrin, F., Pernier, J., Bertrand, O., & Echallier, J. F. '
            '(1989). Spherical splines for scalp potential and current '
            'density mapping. Electroencephalography and Clinical '
            'Neurophysiology, 72(2), 184–187. '
            'https://doi.org/10.1016/0013-4694(89)90180-6'
        )
    else:
        boilerplate.append(
            'No EEG channels were excluded or interpolated based on bad data '
            'quality. '
        )
    
    # Re-referencing
    boilerplate.append(
        'Next, the data were re-referenced to the common average of all EEG '
        'channels.'
    )

    # Ocular correction
    if config['ica_method'] is not None:
        boilerplate.append(
            'Artifacts resulting from blinks and eye movements were corrected '
            'using independent component analysis (ICA). For this, we '
            'temporarily low-pass filtered the data at 1 Hz and extracted the '
            'first 15 independent components using the FastICA algorithm '
            '(Hyvärinen, 1999). Components were then removed automatically '
            'using `find_bads_eog` function in MNE-Python). This function '
            'iteratievly removes components if they are signifcantly '
            'correlated (z > 3.0) with either of two virtual EOG channels '
            '(VEOG: Fp1 - IO1, HEOG: F9 - F10).'
        )
        references.append(
            'Hyvärinen, A. (1999). Fast and robust fixed-point algorithms for '
            'independent component analysis. IEEE Transactions on Neural '
            'Networks, 10(3), 626–634. https://doi.org/10.1109/72.761722'
        )
        
    
    # Combine and print
    boilerplate = ''.join(boilerplate)
    print(boilerplate)
